/*
 * This is part of the ANL/MCS/FL CAVEav distribution.
 *     written by Ivan R. Judson
 *
 */

#include <CAVcam.h>

#ifdef __sgi
 #include <unistd.h>
 #include <strings.h>
#endif

static double matinv(int n, double *ia);

CAVcam::CAVcam(float ex, float ey, float ez, 
		 float cx, float cy, float cz,
		 float ux, float uy, float uz)
{
  _red = 1.0; _green = 0.0; _blue = 0.0;
  _eyex = ex; _eyey = ey; _eyez = ez;
  _centerx = cx; _centery = cy; _centerz = cz;
  _upx = ux; _upy = uy; _upz = uz;

  _captureFrames = 0;

  _modFrames = 1;
}

CAVcam::~CAVcam(void)
{
}

void CAVcam::Init(void)
{
}

void CAVcam::Update(void)
{
  int i;
  float _ViewMat[16];

  if( CAVEEye == CAVE_LEFT_EYE && CAVEMasterDisplay() &&
      !(_fc % _modFrames) && _captureFrames ) 
    {
      // Call the update

      (*UpdateFunc)();

    // Generate the new position/orientation matrix

      glPushMatrix();
      {
	glLoadIdentity();
	gluLookAt(_eyex, _eyey, _eyez, 
		  _eyex + _centerx, _eyey + _centery, _eyez + _centerz, 
		  _eyex + _upx, _eyey + _upy, _eyez + _upz);
	glGetFloatv(GL_MODELVIEW_MATRIX, _ViewMat);
      }
      glPopMatrix();
    
#ifdef DEBUG
      fprintf(stderr, "View Matrix:\n");
      for(i = 0; i < 16; i+=4)
	fprintf(stderr, " %f %f %f %f\n", _ViewMat[i], _ViewMat[i+1],
		_ViewMat[i+2], _ViewMat[i+3]);
      fprintf(stderr, "\n");
#endif

      bcopy(_ViewMat, _invViewMat, sizeof(float) * 16);
      matinv(4, (double *)_invViewMat);

#ifdef DEBUG
      fprintf(stderr, "Inverse View Matrix:\n");
      for(i = 0; i < 16; i+=4)
	fprintf(stderr, " %f %f %f %f\n", _invViewMat[i], _invViewMat[i+1],
		_invViewMat[i+2], _invViewMat[i+3]);
      fprintf(stderr, "\n");
#endif
    }
}


void CAVcam::Draw3DGrid(void)
{
  float x, y, z;

  glPushMatrix();
  {
    glBegin(GL_LINES);
    {
      for(x = -5.0; x < 5.0; x += 1.0)
	for(y = 0.0; y < 10.0; y += 1.0)
	  for(z = -5.0; z < 5.0; z += 1.0)
	    {
	      glColor3f((x+5.0)/10.0, y/10.0, (z+5.0)/10.0);
	    
	      glVertex3f(5.0, y, z);
	      glVertex3f(-5.0, y, z);
	      
	      glVertex3f(x, 0.0, z);
	      glVertex3f(x, 10.0, z);
	      
	      glVertex3f(x, y, 5.0);
	      glVertex3f(x, y, -5.0);
	    }
    }
    glEnd();
  }
  glPopMatrix();
}

void CAVcam::Draw(void)
{
  int i;

  GLfloat cful[] = { -0.25, 0.25, -0.25 };
  GLfloat cfur[] = {  0.25, 0.25, -0.25 };
  GLfloat cfll[] = { -0.25, -0.25, -0.25 };
  GLfloat cflr[] = {  0.25, -0.25, -0.25 };

  GLfloat cbul[] = { -0.45, 0.45, 0.25 };
  GLfloat cbur[] = {  0.45, 0.45, 0.25 };
  GLfloat cbll[] = { -0.45, -0.45, 0.25 };
  GLfloat cblr[] = {  0.45, -0.45, 0.25 };

#ifdef DEBUG
  Draw3DGrid();
#endif

  glPushMatrix();
  {
    glMultMatrixf(_invViewMat);
    glScalef(1.8, 1.1, 1.0);

#ifdef DEBUG
    glBegin(GL_LINES);
    {
      glColor3f(0.0, 1.0, 0.0);
      
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 0.0, -1.0);
      
      glColor3f(1.0, 0.0, 0.0);
      
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 1.0, 0.0);
    }
    glEnd();
#endif
    
#ifndef DEBUG
    glColor3f(_red, _green, _blue);
    
    glBegin(GL_QUAD_STRIP);
    {
      glVertex3fv(cbul);
      glVertex3fv(cful);
      
      glVertex3fv(cbur);
      glVertex3fv(cfur);
      
      glVertex3fv(cblr);
      glVertex3fv(cflr);
      
      glVertex3fv(cbll);
      glVertex3fv(cfll);
      
      glVertex3fv(cbul);
      glVertex3fv(cful);
    }
    glEnd();
#endif

  }
  glPopMatrix();
}

void CAVcam::Process(void)
{
}

void CAVcam::PutPix(void)
{
}

void CAVcam::FlipFrame(unsigned char *data, int Index)
{
  CAVImageDescriptor *image;
  volatile unsigned char *dataTo;
  int i, stride, height, width, depth;

  // Set all pointers to point at the right things
  image = &(_imagepool->_images[Index]);
  dataTo = _imagepool->_images[Index].GetData();

  // retrieve these values once
  height = image->GetHeight();
  width = image->GetWidth();
  depth = image->GetDepth();

  // Calculate the size of one scanline
  stride = width * depth;

#ifdef DEBUG
  fprintf(stderr, "Size < %d %d %d %d >\n", height, width, depth, stride);
#endif
  
  // Loop through and copy each scanline to the correct line in the 
  // output frame buffer
  for(i = 0; i < height; i++)
    bcopy(&(data[i * stride]), (void *)&(dataTo[(height - i) * stride]), 
	  stride);
}



static double matinv(int n, double *ia)
{
  int		i, j, k, iscan, jscan, intch, jtemp;
  int		jcoli, jcolj, jcolk, irowi, irowj, irowk;
  int		irow[4], jcol[4], jord[4];
  double        a[4][4];
  double	y[4], eps, deter;
  double	aijck, pivot;

  /* Return if illegal dimension used */
  if(n <= 0) 
    return(0);

  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      a[i][j] = ia[(i*n) + j];

  /* Set the precision */
  eps = 1.0e-300;
  deter = 1;

  /* Begin elimination procedure */
  for(k=0; k<n; ++k) {
    /* Search for pivot */
    pivot = 0;
    for(i=0; i<n; ++i) {
      /* Scan irow & jcol for invalid subscripts */
      for(j=0; j<n; ++j) {
	for(iscan=0; iscan<k && i != irow[iscan]; ++iscan) {
	  for(jscan=0; jscan<k && j != jcol[jscan]; ++jscan);
	  if(jscan < k)
	    break;
	}
	if(iscan == k && 
	   (a[i][j] < 0 ? -a[i][j] : a[i][j]) > (pivot<0 ? -pivot : pivot)) {
	  pivot = a[i][j];
	  irow[k] = i;
	  jcol[k] = j;
	}
      }
    }
	
    /* Insure pivot is big enough */
    if((pivot < 0 ? -pivot : pivot) <= eps) 
      return(0);

    /* Update determinant value */
    irowk = irow[k];
    jcolk = jcol[k];
    deter *= pivot;

    /* Normalize pivot row */
    for(j=0; j<n; ++j)
      a[irowk][j] /= pivot;
	
    /* Carry out elimination and develop inverse */
    a[irowk][jcolk] = 1 / pivot;
    for(i=0; i<n; ++i) {
      aijck = a[i][jcolk];

      if(i != irowk) {
	a[i][jcolk] = -aijck / pivot;
	for(j=0; j<n; ++j)
	  if(j != jcolk) 
	    a[i][j] -= aijck * a[irowk][j];
      }
    }
  }

  /* Order solution values and create jord array */
  for(i=0; i<n; ++i) {
    irowi = irow[i];
    jcoli = jcol[i];
    jord[irowi] = jcoli;
  }

  /* Adjust sign of determinant */
  intch = 0;
  for(i=0; i<n-1; ++i) {
    for(j=i+1; j<n; ++j) {
      if(jord[j] < jord[i]) {
	jtemp = jord[j];
	jord[j] = jord[i];
	jord[i] = jtemp;
	++intch;
      }
    }
  }

  if(2 * (intch / 2) != intch) 
    deter = -deter;
	
  /* Unscramble the inverse */

  /* First by rows */
  for(j=0; j<n; ++j) {
    for(i=0; i<n; ++i) {
      irowi = irow[i];
      jcoli = jcol[i];
      y[jcoli] = a[irowi][j];
    }
    for(i=0; i<n; ++i) 
      a[i][j] = y[i];
  }
	
  /* Then by columns */
  for(i=0; i<n; ++i) {
    for(j=0; j<n; ++j) {
      irowj = irow[j];
      jcolj = jcol[j];
      y[irowj] = a[i][jcolj];
    }
    for(j=0; j<n; ++j)
      a[i][j] = y[j];
  }

  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      ia[(i*4) + j] = a[i][j];

  return(deter);
}

