/*
 *
 * This is part of the ANL/MCS/FL CAVEav distribution.
 * written by Ivan R. Judson
 *
 */

#include <fileCAVtv.h>

fileCAVtv::fileCAVtv(char *fn, int width, int height, int depth,
		       float x, float y, float z) : CAVtv(x, y, z)
{
#ifdef DEBUG
  fprintf(stderr, "Entering fileCAVtv constructor.\n");
#endif
  _px = x; _py = y; _pz = z;
  _video_width = width; _video_height = height; _video_depth = depth;
  _video_size = _video_width * _video_height * _video_depth;
  _fnbase = fn;

  _s_coord = GLfloat(width) / GLfloat(GetPowTwo(width));
  _t_coord = GLfloat(height) / GLfloat(GetPowTwo(height));

#ifdef DEBUG
  fprintf(stderr, "S < %f > T < %f >\n", _s_coord, _t_coord);
#endif

  _fc = 0;
  _imagepool = new CAVImagePool(width, height, depth, 10);

#ifdef DEBUG
  fprintf(stderr, "Exiting filCAVtv constructor.\n");
#endif
}

fileCAVtv::~fileCAVtv()
{
  delete _imagepool;
}

void fileCAVtv::Init(void)
{
  CAVImageDescriptor *image;
  int nextIndex;

#ifdef DEBUG
  fprintf(stderr, "Entering filCAVtv Initialization.\n");
#endif

  // Get things pointed at the right memory and retrive values
  nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
  image = &(_imagepool->_images[nextIndex]);

#ifdef DEBUG
  fprintf(stderr, "Using image < %d > \n", nextIndex);
#endif

#ifdef DEBUG
  fprintf(stderr, "Creating a texture < %d %d / %d %d >\n", 
	  GetPowTwo(_video_width),
	  GetPowTwo(_video_height),
	  _video_width,
	  _video_height);
#endif

  if(image->GetDepth() == 3)
    glTexImage2D(GL_TEXTURE_2D, 0, 
		 _video_depth, GetPowTwo(_video_width), GetPowTwo(_video_height),
  		 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
  else
    glTexImage2D(GL_TEXTURE_2D, 0,
		 _video_depth, GetPowTwo(_video_width), GetPowTwo(_video_height),
  		 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glEnable(GL_TEXTURE_2D);

#ifdef DEBUG
  fprintf(stderr, "Exiting filCAVtv Initialization.\n");
#endif
}

void fileCAVtv::GetPix(void)
{
  FILE *fp;
  char fn[128];
  CAVImageDescriptor *image;
  int nextIndex;

  sprintf(fn, "%s%d.raw", _fnbase, _fc++);

#ifdef DEBUG
  fprintf(stderr, "Opening %s to read from...", fn);
#endif
  
  nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
  image = &(_imagepool->_images[nextIndex]);

  if((fp = fopen(fn, "r")) != NULL)
    {
      if(fread((void *)image->GetData(), _video_width * _video_height * _video_depth,
	       1, fp))
	{
	  fclose(fp);
#ifdef DEBUG
	  fprintf(stderr, "closed\n");
#endif
	} else {
	  perror("Can't read frame");
	}
    } else {
      perror("Can't open file");
      _fc = 0;
    }
}

void CAVtv::Process(void)
{
  CAVImageDescriptor *image;
  int nextIndex;
  
  // Get things pointed at the right memory and retrive values
  nextIndex = (_imagepool->_nextImage + 1) % _imagepool->_nImages;
  image = &(_imagepool->_images[nextIndex]);

  if(image->GetDepth() == 3)
    glTexSubImage2DEXT(GL_TEXTURE_2D, 0, 0, 0, image->GetWidth(),
		       image->GetHeight(), GL_RGB, GL_UNSIGNED_BYTE, 
		       (void *)image->GetData());
  else
    glTexSubImage2DEXT(GL_TEXTURE_2D, 0, 0, 0, image->GetWidth(),
		       image->GetHeight(), GL_RGBA, GL_UNSIGNED_BYTE, 
		       (void *)image->GetData());
}
